@@ -21,6 +21,7 @@
 import logging
 import multiprocessing
 import os
+import re
 import subprocess
 import sys
 import shutil
@@ -33,6 +34,7 @@ import click
 import gi
 
 from .. import _signals  # nopep8
+from .._profile import Topics, profile_start, profile_end
 
 gi.require_version('OSTree', '1.0')
 # pylint: disable=wrong-import-position,wrong-import-order
@@ -251,7 +253,7 @@ class PushMessageReader(object):
         if not header:
             # Remote end quit
             return None, None
-        order, version, cmdtype, size = self.decode_header(header)
+        order, _, cmdtype, size = self.decode_header(header)
         msg = self.file.read(size)
         if len(msg) != size:
             raise PushException('Did not receive full message')
@@ -268,11 +270,11 @@ class PushMessageReader(object):
         return cmdtype, args.unpack()
 
     def receive_info(self):
-        cmdtype, args = self.receive([PushCommandType.info])
+        _, args = self.receive([PushCommandType.info])
         return args
 
     def receive_update(self):
-        cmdtype, args = self.receive([PushCommandType.update])
+        _, args = self.receive([PushCommandType.update])
         return args
 
     def receive_putobjects(self, repo):
@@ -307,11 +309,11 @@ class PushMessageReader(object):
         return received_objects
 
     def receive_status(self):
-        cmdtype, args = self.receive([PushCommandType.status])
+        _, args = self.receive([PushCommandType.status])
         return args
 
     def receive_done(self):
-        cmdtype, args = self.receive([PushCommandType.done])
+        _, args = self.receive([PushCommandType.done])
         return args
 
 
@@ -387,6 +389,9 @@ class ProcessWithPipes(object):
             self.stderr = open(r2, 'rb')
             os.close(w2)
 
+        # The eventual return code
+        self.returncode = -1
+
     def wait(self):
         self.proc.join()
         self.returncode = self.proc.exitcode
@@ -616,6 +621,16 @@ class OSTreeReceiver(object):
             return 0
         update_refs = args
 
+        profile_names = set()
+        for update_ref in update_refs:
+            # Strip off the SHA256 sum on the right of the reference,
+            # leaving the project and element name
+            project_and_element_name = re.sub(r"/[a-z0-9]+$", '', update_ref)
+            profile_names.add(project_and_element_name)
+
+        profile_name = '_'.join(profile_names)
+        profile_start(Topics.ARTIFACT_RECEIVE, profile_name)
+
         self.writer.send_status(True)
 
         # Wait for putobjects or done
@@ -658,6 +673,8 @@ class OSTreeReceiver(object):
         # Inform pusher that everything is in place
         self.writer.send_done()
 
+        profile_end(Topics.ARTIFACT_RECEIVE, profile_name)
+
         return 0
 
 
