@@ -29,7 +29,13 @@ _last_task_error_domain = None
 _last_task_error_reason = None
 
 
-def _get_last_exception():
+# get_last_exception()
+#
+# Fetches the last exception from the main process
+#
+# Used by regression tests
+#
+def get_last_exception():
     global _last_exception
 
     le = _last_exception
@@ -37,8 +43,13 @@ def _get_last_exception():
     return le
 
 
-# Called from regression test fixtures
-def _get_last_task_error():
+# get_last_task_error()
+#
+# Fetches the last exception from a task
+#
+# Used by regression tests
+#
+def get_last_task_error():
     global _last_task_error_domain
     global _last_task_error_reason
 
@@ -48,8 +59,14 @@ def _get_last_task_error():
     return (d, r)
 
 
-# Called from the scheduler when an error is encountered
-def _set_last_task_error(domain, reason):
+# set_last_task_error()
+#
+# Sets the last exception of a task
+#
+# This is set by some internals to inform regression
+# tests about how things failed in a machine readable way
+#
+def set_last_task_error(domain, reason):
     global _last_task_error_domain
     global _last_task_error_reason
 
@@ -70,6 +87,7 @@ class ErrorDomain(Enum):
     PROG_NOT_FOUND = 12
     SOURCE = 10
     ELEMENT = 11
+    APP = 12
 
 
 # BstError is an internal base exception class for BuildSream
@@ -149,17 +167,35 @@ class LoadErrorReason(Enum):
     # BuildStream does not support the required project format version
     UNSUPPORTED_PROJECT = 7
 
+    # Project requires a newer version of a plugin than the one which was loaded
+    UNSUPPORTED_PLUGIN = 8
+
     # A conditional expression failed to resolve
-    EXPRESSION_FAILED = 8
+    EXPRESSION_FAILED = 9
 
     # An assertion was intentionally encoded into project YAML
-    USER_ASSERTION = 9
+    USER_ASSERTION = 10
 
     # A list composition directive did not apply to any underlying list
-    TRAILING_LIST_DIRECTIVE = 10
+    TRAILING_LIST_DIRECTIVE = 11
 
     # Conflicting junctions in subprojects
-    CONFLICTING_JUNCTION = 11
+    CONFLICTING_JUNCTION = 12
+
+    # Failure to load a project from a specified junction
+    INVALID_JUNCTION = 13
+
+    # Subproject needs to be fetched
+    SUBPROJECT_FETCH_NEEDED = 14
+
+    # Subproject has no ref
+    SUBPROJECT_INCONSISTENT = 15
+
+    # An invalid symbol name was encountered
+    INVALID_SYMBOL_NAME = 16
+
+    # A project.conf file was missing
+    MISSING_PROJECT_CONF = 17
 
 
 # LoadError
@@ -174,8 +210,8 @@ class LoadErrorReason(Enum):
 # interpreting project YAML
 #
 class LoadError(BstError):
-    def __init__(self, reason, message):
-        super().__init__(message, domain=ErrorDomain.LOAD, reason=reason)
+    def __init__(self, reason, message, *, detail=None):
+        super().__init__(message, detail=detail, domain=ErrorDomain.LOAD, reason=reason)
 
 
 # ImplError
@@ -220,7 +256,7 @@ class ArtifactError(BstError):
 #
 class PipelineError(BstError):
 
-    def __init__(self, message=None, *, detail=None, reason=None):
+    def __init__(self, message=None, *, detail=None, reason=None, terminated=False):
 
         # The empty string should never appear to a user,
         # this only allows us to treat this internal error as
@@ -229,3 +265,14 @@ class PipelineError(BstError):
             message = ""
 
         super().__init__(message, detail=detail, domain=ErrorDomain.PIPELINE, reason=reason)
+
+        self.terminated = terminated
+
+
+# AppError
+#
+# Raised from the frontend App directly
+#
+class AppError(BstError):
+    def __init__(self, message, detail=None, reason=None):
+        super().__init__(message, detail=detail, domain=ErrorDomain.APP, reason=reason)
