@@ -3,8 +3,8 @@ import sys
 
 import click
 from .. import _yaml
-from .._exceptions import BstError, PipelineError, LoadError
-from ..__version__ import __version__ as build_stream_version
+from .._exceptions import BstError, LoadError
+from .._versions import BST_FORMAT_VERSION
 from .complete import main_bashcomplete, complete_path, CompleteUnhandled
 
 
@@ -111,13 +111,23 @@ click.BaseCommand.main = override_main
 ##################################################################
 #                          Main Options                          #
 ##################################################################
+def print_version(ctx, param, value):
+    if not value or ctx.resilient_parsing:
+        return
+
+    from .. import __version__
+    click.echo(__version__)
+    ctx.exit()
+
+
 @click.group(context_settings=dict(help_option_names=['-h', '--help']))
-@click.version_option(version=build_stream_version)
+@click.option('--version', is_flag=True, callback=print_version,
+              expose_value=False, is_eager=True)
 @click.option('--config', '-c',
               type=click.Path(exists=True, dir_okay=False, readable=True),
               help="Configuration file to use")
 @click.option('--directory', '-C', default=os.getcwd(),
-              type=click.Path(exists=True, file_okay=False, readable=True),
+              type=click.Path(file_okay=False, readable=True),
               help="Project directory (default: current directory)")
 @click.option('--on-error', default=None,
               type=click.Choice(['continue', 'quit', 'terminate']),
@@ -157,13 +167,38 @@ def cli(context, **kwargs):
     user preferences configuration file.
     """
 
-    from .main import App
+    from .app import App
 
     # Create the App, giving it the main arguments
     context.obj = App(dict(kwargs))
     context.call_on_close(context.obj.cleanup)
 
 
+##################################################################
+#                           Init Command                         #
+##################################################################
+@cli.command(short_help="Initialize a new BuildStream project")
+@click.option('--project-name', type=click.STRING,
+              help="The project name to use")
+@click.option('--format-version', type=click.INT, default=BST_FORMAT_VERSION,
+              help="The required format version (default: {})".format(BST_FORMAT_VERSION))
+@click.option('--element-path', type=click.Path(), default="elements",
+              help="The subdirectory to store elements in (default: elements)")
+@click.option('--force', '-f', default=False, is_flag=True,
+              help="Allow overwriting an existing project.conf")
+@click.pass_obj
+def init(app, project_name, format_version, element_path, force):
+    """Initialize a new BuildStream project
+
+    Creates a new BuildStream project.conf in the project
+    directory.
+
+    Unless `--project-name` is specified, this will be an
+    interactive session.
+    """
+    app.init_project(project_name, format_version, element_path, force)
+
+
 ##################################################################
 #                          Build Command                         #
 ##################################################################
@@ -179,16 +214,19 @@ def cli(context, **kwargs):
 @click.option('--track-except', multiple=True,
               type=click.Path(dir_okay=False, readable=True),
               help="Except certain dependencies from tracking")
+@click.option('--track-cross-junctions', '-J', default=False, is_flag=True,
+              help="Allow tracking to cross junction boundaries")
 @click.option('--track-save', default=False, is_flag=True,
               help="Deprecated: This is ignored")
 @click.argument('elements', nargs=-1,
                 type=click.Path(dir_okay=False, readable=True))
 @click.pass_obj
-def build(app, elements, all_, track_, track_save, track_all, track_except):
+def build(app, elements, all_, track_, track_save, track_all, track_except, track_cross_junctions):
     """Build elements in a pipeline"""
 
-    if track_except and not (track_ or track_all):
-        click.echo("ERROR: --track-except cannot be used without --track or --track-all", err=True)
+    if (track_except or track_cross_junctions) and not (track_ or track_all):
+        click.echo("ERROR: The --track-except and --track-cross-junctions options "
+                   "can only be used with --track or --track-all", err=True)
         sys.exit(-1)
 
     if track_save:
@@ -201,17 +239,11 @@ def build(app, elements, all_, track_, track_save, track_all, track_except):
     if track_:
         rewritable = True
 
-    app.initialize(elements, except_=track_except, rewritable=rewritable,
-                   use_configured_remote_caches=True, track_elements=track_,
-                   fetch_subprojects=True)
-    app.print_heading()
-    try:
-        app.pipeline.build(app.scheduler, all_, track_)
-        app.print_summary()
-    except PipelineError as e:
-        app.print_error(e)
-        app.print_summary()
-        sys.exit(-1)
+    with app.initialized(elements, session_name="Build", except_=track_except, rewritable=rewritable,
+                         use_configured_remote_caches=True, track_elements=track_,
+                         track_cross_junctions=track_cross_junctions,
+                         fetch_subprojects=True):
+        app.pipeline.build(app.scheduler, build_all=all_)
 
 
 ##################################################################
@@ -226,10 +258,12 @@ def build(app, elements, all_, track_, track_save, track_all, track_except):
               help='The dependencies to fetch (default: plan)')
 @click.option('--track', 'track_', default=False, is_flag=True,
               help="Track new source references before fetching")
+@click.option('--track-cross-junctions', '-J', default=False, is_flag=True,
+              help="Allow tracking to cross junction boundaries")
 @click.argument('elements', nargs=-1,
                 type=click.Path(dir_okay=False, readable=True))
 @click.pass_obj
-def fetch(app, elements, deps, track_, except_):
+def fetch(app, elements, deps, track_, except_, track_cross_junctions):
     """Fetch sources required to build the pipeline
 
     By default this will only try to fetch sources which are
@@ -244,20 +278,17 @@ def fetch(app, elements, deps, track_, except_):
         plan:  Only dependencies required for the build plan
         all:   All dependencies
     """
-
-    app.initialize(elements, except_=except_, rewritable=track_,
-                   track_elements=elements if track_ else None,
-                   fetch_subprojects=True)
-    try:
-        dependencies = app.pipeline.deps_elements(deps)
-        app.print_heading(deps=dependencies)
-        app.pipeline.fetch(app.scheduler, dependencies, track_)
-        app.print_summary()
-    except PipelineError as e:
-        app.print_error(e)
-        app.print_summary()
+    if track_cross_junctions and not track_:
+        click.echo("ERROR: The --track-cross-junctions option can only be used with --track", err=True)
         sys.exit(-1)
 
+    with app.initialized(elements, session_name="Fetch", except_=except_, rewritable=track_,
+                         track_elements=elements if track_ else None,
+                         track_cross_junctions=track_cross_junctions,
+                         fetch_subprojects=True):
+        dependencies = app.pipeline.get_selection(deps)
+        app.pipeline.fetch(app.scheduler, dependencies)
+
 
 ##################################################################
 #                          Track Command                         #
@@ -269,10 +300,12 @@ def fetch(app, elements, deps, track_, except_):
 @click.option('--deps', '-d', default='none',
               type=click.Choice(['none', 'all']),
               help='The dependencies to track (default: none)')
+@click.option('--cross-junctions', '-J', default=False, is_flag=True,
+              help="Allow crossing junction boundaries")
 @click.argument('elements', nargs=-1,
                 type=click.Path(dir_okay=False, readable=True))
 @click.pass_obj
-def track(app, elements, deps, except_):
+def track(app, elements, deps, except_, cross_junctions):
     """Consults the specified tracking branches for new versions available
     to build and updates the project with any newly available references.
 
@@ -282,20 +315,15 @@ def track(app, elements, deps, except_):
     Specify `--deps` to control which sources to track:
 
     \b
-        none:  No dependencies, just the element itself
-        all:   All dependencies
+        none:  No dependencies, just the specified elements
+        all:   All dependencies of all specified elements
     """
-    app.initialize(elements, except_=except_, rewritable=True, track_elements=elements,
-                   fetch_subprojects=True)
-    try:
-        dependencies = app.pipeline.deps_elements(deps)
-        app.print_heading(deps=dependencies)
-        app.pipeline.track(app.scheduler, dependencies)
-        app.print_summary()
-    except PipelineError as e:
-        app.print_error(e)
-        app.print_summary()
-        sys.exit(-1)
+    with app.initialized(elements, session_name="Track", except_=except_, rewritable=True,
+                         track_elements=elements,
+                         track_cross_junctions=cross_junctions,
+                         track_selection=deps,
+                         fetch_subprojects=True):
+        app.pipeline.track(app.scheduler)
 
 
 ##################################################################
@@ -323,16 +351,10 @@ def pull(app, elements, deps, remote):
         none:  No dependencies, just the element itself
         all:   All dependencies
     """
-    app.initialize(elements, use_configured_remote_caches=(remote is None),
-                   add_remote_cache=remote, fetch_subprojects=True)
-    try:
-        to_pull = app.pipeline.deps_elements(deps)
+    with app.initialized(elements, session_name="Pull", use_configured_remote_caches=(remote is None),
+                         add_remote_cache=remote, fetch_subprojects=True):
+        to_pull = app.pipeline.get_selection(deps)
         app.pipeline.pull(app.scheduler, to_pull)
-        app.print_summary()
-    except BstError as e:
-        app.print_error(e)
-        app.print_summary()
-        sys.exit(-1)
 
 
 ##################################################################
@@ -359,16 +381,11 @@ def push(app, elements, deps, remote):
         none:  No dependencies, just the element itself
         all:   All dependencies
     """
-    app.initialize(elements, use_configured_remote_caches=(remote is None),
-                   add_remote_cache=remote, fetch_subprojects=True)
-    try:
-        to_push = app.pipeline.deps_elements(deps)
+    with app.initialized(elements, session_name="Push",
+                         use_configured_remote_caches=(remote is None),
+                         add_remote_cache=remote, fetch_subprojects=True):
+        to_push = app.pipeline.get_selection(deps)
         app.pipeline.push(app.scheduler, to_push)
-        app.print_summary()
-    except BstError as e:
-        app.print_error(e)
-        app.print_summary()
-        sys.exit(-1)
 
 
 ##################################################################
@@ -439,20 +456,17 @@ def show(app, elements, deps, except_, order, format_, downloadable):
         bst show target.bst --format \\
             $'---------- %{name} ----------\\n%{vars}'
     """
-    app.initialize(elements, except_=except_, use_configured_remote_caches=downloadable)
-    try:
-        dependencies = app.pipeline.deps_elements(deps)
-    except PipelineError as e:
-        click.echo("{}".format(e), err=True)
-        sys.exit(-1)
+    with app.initialized(elements, except_=except_, use_configured_remote_caches=downloadable):
 
-    if order == "alpha":
-        dependencies = sorted(dependencies)
+        dependencies = app.pipeline.get_selection(deps)
+        if order == "alpha":
+            dependencies = sorted(dependencies)
 
-    if not format_:
-        format_ = app.context.log_element_format
+        if not format_:
+            format_ = app.context.log_element_format
+
+        report = app.logger.show_pipeline(dependencies, format_)
 
-    report = app.logger.show_pipeline(dependencies, format_)
     click.echo(report, color=app.colors)
 
 
@@ -498,7 +512,8 @@ def shell(app, element, sysroot, mount, isolate, build_, command):
     else:
         scope = Scope.RUN
 
-    app.initialize((element,))
+    with app.initialized((element,)):
+        pass
 
     # Assert we have everything we need built.
     missing_deps = []
@@ -548,13 +563,8 @@ def shell(app, element, sysroot, mount, isolate, build_, command):
 def checkout(app, element, directory, force, integrate, hardlinks):
     """Checkout a built artifact to the specified directory
     """
-    app.initialize((element,))
-    try:
+    with app.initialized((element,)):
         app.pipeline.checkout(directory, force, integrate, hardlinks)
-        click.echo("", err=True)
-    except BstError as e:
-        app.print_error(e)
-        sys.exit(-1)
 
 
 ##################################################################
@@ -578,18 +588,12 @@ def checkout(app, element, directory, force, integrate, hardlinks):
 @click.pass_obj
 def source_bundle(app, target, force, directory,
                   track_, compression, except_):
-    """Produce a source bundle to be manually executed"""
-    app.initialize((target,), rewritable=track_, track_elements=[target] if track_ else None)
-    try:
-        dependencies = app.pipeline.deps_elements('all')
-        app.print_heading(dependencies)
+    """Produce a source bundle to be manually executed
+    """
+    with app.initialized((target,), rewritable=track_, track_elements=[target] if track_ else None):
+        dependencies = app.pipeline.get_selection('all')
         app.pipeline.source_bundle(app.scheduler, dependencies, force, track_,
                                    compression, directory)
-        click.echo("", err=True)
-    except BstError as e:
-        click.echo("", err=True)
-        click.echo("ERROR: {}".format(e), err=True)
-        sys.exit(-1)
 
 
 ##################################################################
@@ -618,45 +622,45 @@ def workspace():
 def workspace_open(app, no_checkout, force, track_, element, directory):
     """Open a workspace for manual source modification"""
 
-    app.initialize((element,), rewritable=track_, track_elements=[element] if track_ else None)
-    try:
-        app.pipeline.open_workspace(app.scheduler, directory, no_checkout, track_, force)
-        click.echo("", err=True)
-    except BstError as e:
-        app.print_error(e)
-        sys.exit(-1)
+    if os.path.exists(directory):
+
+        if not os.path.isdir(directory):
+            click.echo("Checkout directory is not a directory: {}".format(directory), err=True)
+            sys.exit(-1)
+
+        if not (no_checkout or force) and os.listdir(directory):
+            click.echo("Checkout directory is not empty: {}".format(directory), err=True)
+            sys.exit(-1)
+
+    with app.initialized((element,), rewritable=track_, track_elements=[element] if track_ else None):
+        # This command supports only one target
+        target = app.pipeline.targets[0]
+        app.open_workspace(target, directory, no_checkout, track_, force)
 
 
 ##################################################################
 #                     Workspace Close Command                    #
 ##################################################################
-@workspace.command(name='close', short_help="Close a workspace")
+@workspace.command(name='close', short_help="Close workspaces")
 @click.option('--remove-dir', default=False, is_flag=True,
               help="Remove the path that contains the closed workspace")
-@click.argument('element',
+@click.option('--all', '-a', 'all_', default=False, is_flag=True,
+              help="Close all open workspaces")
+@click.argument('elements', nargs=-1,
                 type=click.Path(dir_okay=False, readable=True))
 @click.pass_obj
-def workspace_close(app, remove_dir, element):
+def workspace_close(app, remove_dir, all_, elements):
     """Close a workspace"""
 
-    app.initialize((element,))
-
-    if app.pipeline.project._get_workspace(app.pipeline.targets[0].name) is None:
-        click.echo("ERROR: Workspace '{}' does not exist".format(element), err=True)
+    if not (all_ or elements):
+        click.echo('ERROR: no elements specified', err=True)
         sys.exit(-1)
 
-    if app.interactive and remove_dir:
-        if not click.confirm('This will remove all your changes, are you sure?'):
-            click.echo('Aborting', err=True)
-            sys.exit(-1)
-
-    try:
-        app.pipeline.close_workspace(remove_dir)
-        click.echo("", err=True)
-    except BstError as e:
-        click.echo("", err=True)
-        click.echo("ERROR: {}".format(e), err=True)
-        sys.exit(-1)
+    with app.partially_initialized():
+        if all_:
+            elements = [element_name for element_name, _ in app.project.workspaces.list()]
+        for element in elements:
+            app.close_workspace(element, remove_dir)
 
 
 ##################################################################
@@ -665,26 +669,30 @@ def workspace_close(app, remove_dir, element):
 @workspace.command(name='reset', short_help="Reset a workspace to its original state")
 @click.option('--track', 'track_', default=False, is_flag=True,
               help="Track and fetch the latest source before resetting")
-@click.option('--no-checkout', default=False, is_flag=True,
-              help="Do not checkout the source, only link to the given directory")
-@click.argument('element',
+@click.option('--all', '-a', 'all_', default=False, is_flag=True,
+              help="Reset all open workspaces")
+@click.argument('elements', nargs=-1,
                 type=click.Path(dir_okay=False, readable=True))
 @click.pass_obj
-def workspace_reset(app, track_, no_checkout, element):
+def workspace_reset(app, track_, all_, elements):
     """Reset a workspace to its original state"""
-    app.initialize((element,))
+
+    if not (all_ or elements):
+        click.echo('ERROR: no elements specified', err=True)
+        sys.exit(-1)
+
     if app.interactive:
         if not click.confirm('This will remove all your changes, are you sure?'):
             click.echo('Aborting', err=True)
             sys.exit(-1)
 
-    try:
-        app.pipeline.reset_workspace(app.scheduler, track_, no_checkout)
-        click.echo("", err=True)
-    except BstError as e:
-        click.echo("", err=True)
-        click.echo("ERROR: {}".format(e), err=True)
-        sys.exit(-1)
+    with app.partially_initialized():
+        if all_:
+            elements = tuple(element_name for element_name, _ in app.project.workspaces.list())
+
+    with app.initialized(elements):
+        for target in app.pipeline.targets:
+            app.reset_workspace(target, track_)
 
 
 ##################################################################
@@ -695,35 +703,15 @@ def workspace_reset(app, track_, no_checkout, element):
 def workspace_list(app):
     """List open workspaces"""
 
-    from .. import _yaml
-    from .._context import Context
-    from .._project import Project
+    with app.partially_initialized():
+        workspaces = []
+        for element_name, workspace_ in app.project.workspaces.list():
+            workspace_detail = {
+                'element': element_name,
+                'directory': workspace_.path,
+            }
+            workspaces.append(workspace_detail)
 
-    directory = app.main_options['directory']
-    config = app.main_options['config']
-
-    try:
-        context = Context()
-        context.load(config)
-    except BstError as e:
-        click.echo("Error loading user configuration: {}".format(e), err=True)
-        sys.exit(-1)
-
-    try:
-        project = Project(directory, context, cli_options=app.main_options['option'])
-    except BstError as e:
-        click.echo("Error loading project: {}".format(e), err=True)
-        sys.exit(-1)
-
-    workspaces = []
-    for element_name, directory in project._list_workspaces():
-        workspace_ = {
-            'element': element_name,
-            'directory': directory,
-        }
-
-        workspaces.append(workspace_)
-
-    _yaml.dump({
-        'workspaces': workspaces
-    })
+        _yaml.dump({
+            'workspaces': workspaces
+        })
