@@ -39,7 +39,7 @@ initialized = False
 #
 #   BST_PROFILE=circ-dep-check:sort-deps bst <command> <args>
 #
-# The special 'all' value will enable all profiles
+# The special 'all' value will enable all profiles.
 class Topics():
     CIRCULAR_CHECK = 'circ-dep-check'
     SORT_DEPENDENCIES = 'sort-deps'
@@ -48,6 +48,7 @@ class Topics():
     LOAD_PROJECT = 'load-project'
     LOAD_PIPELINE = 'load-pipeline'
     SHOW = 'show'
+    ARTIFACT_RECEIVE = 'artifact-receive'
     ALL = 'all'
 
 
