@@ -33,16 +33,10 @@ from ._artifactcache import artifact_cache_specs_from_config_node
 from ._elementfactory import ElementFactory
 from ._sourcefactory import SourceFactory
 from ._projectrefs import ProjectRefs, ProjectRefStorage
+from ._versions import BST_FORMAT_VERSION
+from ._workspaces import Workspaces
 
 
-# The base BuildStream format version
-#
-# This version is bumped whenever enhancements are made
-# to the `project.conf` format or the core element format.
-#
-BST_FORMAT_VERSION = 6
-BST_WORKSPACE_FORMAT_VERSION = 1
-
 # The separator we use for user specified aliases
 _ALIAS_SEPARATOR = ':'
 
@@ -86,28 +80,34 @@ class Project():
         # Absolute path to where elements are loaded from within the project
         self.element_path = None
 
+        # ProjectRefs for the main refs and also for junctions
+        self.refs = ProjectRefs(self.directory, 'project.refs')
+        self.junction_refs = ProjectRefs(self.directory, 'junction.refs')
+
+        self.workspaces = None                   # Workspaces
+        self.options = None                      # OptionPool
+        self.junction = junction                 # The junction Element object, if this is a subproject
+        self.fail_on_overlap = False             # Whether overlaps are treated as errors
+        self.ref_storage = None                  # ProjectRefStorage setting
+        self.base_variables = {}                 # The base set of variables
+        self.base_environment = {}               # The base set of environment variables
+        self.base_env_nocache = None             # The base nocache mask (list) for the environment
+        self.element_overrides = {}              # Element specific configurations
+        self.source_overrides = {}               # Source specific configurations
+
+        #
+        # Private Members
+        #
         self._context = context  # The invocation Context
-        self._variables = {}     # The default variables overridden with project wide overrides
-        self._environment = {}   # The base sandbox environment
-        self._elements = {}      # Element specific configurations
-        self._sources = {}       # Source specific configurations
         self._aliases = {}       # Aliases dictionary
-        self._workspaces = {}    # Workspaces
         self._plugin_source_origins = []   # Origins of custom sources
         self._plugin_element_origins = []  # Origins of custom elements
-        self._options = None    # Project options, the OptionPool
-        self._junction = junction   # The junction element, if this is a subproject
+
         self._cli_options = cli_options
         self._cache_key = None
         self._source_format_versions = {}
         self._element_format_versions = {}
-        self._fail_on_overlap = False
-        self._ref_storage = None  # The ProjectRefStorage setting
 
-        # The project.refs management object for this project
-        self.refs = ProjectRefs(self.directory)
-
-        # Shell options
         self._shell_command = []      # The default interactive shell command
         self._shell_environment = {}  # Statically set environment vars
         self._shell_host_files = []   # A list of HostMount objects
@@ -116,7 +116,7 @@ class Project():
         self._load()
         profile_end(Topics.LOAD_PROJECT, self.directory.replace(os.sep, '-'))
 
-        self._context._add_project(self)
+        self._context.add_project(self)
 
     # translate_url():
     #
@@ -141,6 +141,67 @@ class Project():
 
         return url
 
+    # get_shell_config()
+    #
+    # Gets the project specified shell configuration
+    #
+    # Returns:
+    #    (list): The shell command
+    #    (dict): The shell environment
+    #    (list): The list of HostMount objects
+    #
+    def get_shell_config(self):
+        return (self._shell_command, self._shell_environment, self._shell_host_files)
+
+    # get_cache_key():
+    #
+    # Returns the cache key, calculating it if necessary
+    #
+    # Returns:
+    #    (str): A hex digest cache key for the Context
+    #
+    def get_cache_key(self):
+        if self._cache_key is None:
+
+            # Anything that alters the build goes into the unique key
+            # (currently nothing here)
+            self._cache_key = _cachekey.generate_key({})
+
+        return self._cache_key
+
+    # create_element()
+    #
+    # Instantiate and return an element
+    #
+    # Args:
+    #    artifacts (ArtifactCache): The artifact cache
+    #    meta (MetaElement): The loaded MetaElement
+    #
+    # Returns:
+    #    (Element): A newly created Element object of the appropriate kind
+    #
+    def create_element(self, artifacts, meta):
+        element = self._element_factory.create(self._context, self, artifacts, meta)
+        version = self._element_format_versions.get(meta.kind, 0)
+        self._assert_plugin_format(element, version)
+        return element
+
+    # create_source()
+    #
+    # Instantiate and return a Source
+    #
+    # Args:
+    #    meta (MetaSource): The loaded MetaSource
+    #
+    # Returns:
+    #    (Source): A newly created Source object of the appropriate kind
+    #
+    def create_source(self, meta):
+        source = self._source_factory.create(self._context, self, meta)
+        version = self._source_format_versions.get(meta.kind, 0)
+        self._assert_plugin_format(source, version)
+        return source
+
     # _load():
     #
     # Loads the project configuration file in the project directory.
@@ -154,17 +215,32 @@ class Project():
         config = _yaml.load(_site.default_project_config)
 
         # Load project local config and override the builtin
-        project_conf = _yaml.load(projectfile)
+        try:
+            project_conf = _yaml.load(projectfile)
+        except LoadError as e:
+            # Raise a more specific error here
+            raise LoadError(LoadErrorReason.MISSING_PROJECT_CONF, str(e))
+
         _yaml.composite(config, project_conf)
 
         # Element and Source  type configurations will be composited later onto
         # element/source types, so we delete it from here and run our final
         # assertion after.
-        self._elements = _yaml.node_get(config, Mapping, 'elements', default_value={})
-        self._sources = _yaml.node_get(config, Mapping, 'sources', default_value={})
+        self.element_overrides = _yaml.node_get(config, Mapping, 'elements', default_value={})
+        self.source_overrides = _yaml.node_get(config, Mapping, 'sources', default_value={})
         config.pop('elements', None)
         config.pop('sources', None)
         _yaml.node_final_assertions(config)
+
+        # Assert project's format version early, before validating toplevel keys
+        format_version = _yaml.node_get(config, int, 'format-version')
+        if BST_FORMAT_VERSION < format_version:
+            major, minor = utils.get_bst_version()
+            raise LoadError(
+                LoadErrorReason.UNSUPPORTED_PROJECT,
+                "Project requested format version {}, but BuildStream {}.{} only supports up until format version {}"
+                .format(format_version, major, minor, BST_FORMAT_VERSION))
+
         _yaml.node_validate(config, [
             'format-version',
             'element-path', 'variables',
@@ -179,6 +255,11 @@ class Project():
         # The project name, element path and option declarations
         # are constant and cannot be overridden by option conditional statements
         self.name = _yaml.node_get(config, str, 'name')
+
+        # Validate that project name is a valid symbol name
+        _yaml.assert_symbol_name(_yaml.node_get_provenance(config, 'name'),
+                                 self.name, "project name")
+
         self.element_path = os.path.join(
             self.directory,
             _yaml.node_get(config, str, 'element-path')
@@ -186,28 +267,28 @@ class Project():
 
         # Load project options
         options_node = _yaml.node_get(config, Mapping, 'options', default_value={})
-        self._options = OptionPool(self.element_path)
-        self._options.load(options_node)
-        if self._junction:
+        self.options = OptionPool(self.element_path)
+        self.options.load(options_node)
+        if self.junction:
             # load before user configuration
-            self._options.load_yaml_values(self._junction.options, transform=self._junction._subst_string)
+            self.options.load_yaml_values(self.junction.options, transform=self.junction._subst_string)
 
         # Collect option values specified in the user configuration
-        overrides = self._context._get_overrides(self.name)
+        overrides = self._context.get_overrides(self.name)
         override_options = _yaml.node_get(overrides, Mapping, 'options', default_value={})
-        self._options.load_yaml_values(override_options)
+        self.options.load_yaml_values(override_options)
         if self._cli_options:
-            self._options.load_cli_values(self._cli_options)
+            self.options.load_cli_values(self._cli_options)
 
         # We're done modifying options, now we can use them for substitutions
-        self._options.resolve()
+        self.options.resolve()
 
         #
         # Now resolve any conditionals in the remaining configuration,
         # any conditionals specified for project option declarations,
         # or conditionally specifying the project name; will be ignored.
         #
-        self._options.process_node(config)
+        self.options.process_node(config)
 
         #
         # Now all YAML composition is done, from here on we just load
@@ -218,17 +299,7 @@ class Project():
         self.artifact_cache_specs = artifact_cache_specs_from_config_node(config)
 
         # Workspace configurations
-        workspace_config = self._load_workspace_config()
-        self._workspaces = self._parse_workspace_config(workspace_config)
-
-        # Assert project version
-        format_version = _yaml.node_get(config, int, 'format-version')
-        if BST_FORMAT_VERSION < format_version:
-            major, minor = utils.get_bst_version()
-            raise LoadError(
-                LoadErrorReason.UNSUPPORTED_PROJECT,
-                "Project requested format version {}, but BuildStream {}.{} only supports up until format version {}"
-                .format(format_version, major, minor, BST_FORMAT_VERSION))
+        self.workspaces = Workspaces(self)
 
         # Plugin origins and versions
         origins = _yaml.node_get(config, list, 'plugins', default_value=[])
@@ -278,23 +349,21 @@ class Project():
         self._aliases = _yaml.node_get(config, Mapping, 'aliases', default_value={})
 
         # Load base variables
-        self._variables = _yaml.node_get(config, Mapping, 'variables')
+        self.base_variables = _yaml.node_get(config, Mapping, 'variables')
 
         # Add the project name as a default variable
-        self._variables['project-name'] = self.name
+        self.base_variables['project-name'] = self.name
 
         # Extend variables with automatic variables and option exports
         # Initialize it as a string as all variables are processed as strings.
-        self._variables['max-jobs'] = str(multiprocessing.cpu_count())
+        self.base_variables['max-jobs'] = str(multiprocessing.cpu_count())
 
         # Export options into variables, if that was requested
-        for _, option in self._options.options.items():
-            if option.variable:
-                self._variables[option.variable] = option.get_value()
+        self.options.export_variables(self.base_variables)
 
         # Load sandbox environment variables
-        self._environment = _yaml.node_get(config, Mapping, 'environment')
-        self._env_nocache = _yaml.node_get(config, list, 'environment-nocache')
+        self.base_environment = _yaml.node_get(config, Mapping, 'environment')
+        self.base_env_nocache = _yaml.node_get(config, list, 'environment-nocache')
 
         # Load sandbox configuration
         self._sandbox = _yaml.node_get(config, Mapping, 'sandbox')
@@ -303,19 +372,20 @@ class Project():
         self._splits = _yaml.node_get(config, Mapping, 'split-rules')
 
         # Fail on overlap
-        self._fail_on_overlap = _yaml.node_get(config, bool, 'fail-on-overlap')
+        self.fail_on_overlap = _yaml.node_get(config, bool, 'fail-on-overlap')
 
         # Use separate file for storing source references
-        self._ref_storage = _yaml.node_get(config, str, 'ref-storage')
-        if self._ref_storage not in [ProjectRefStorage.INLINE, ProjectRefStorage.PROJECT_REFS]:
+        self.ref_storage = _yaml.node_get(config, str, 'ref-storage')
+        if self.ref_storage not in [ProjectRefStorage.INLINE, ProjectRefStorage.PROJECT_REFS]:
             p = _yaml.node_get_provenance(config, 'ref-storage')
             raise LoadError(LoadErrorReason.INVALID_DATA,
                             "{}: Invalid value '{}' specified for ref-storage"
-                            .format(p, self._ref_storage))
+                            .format(p, self.ref_storage))
 
         # Load project.refs if it exists, this may be ignored.
-        if self._ref_storage == ProjectRefStorage.PROJECT_REFS:
-            self.refs.load(self._options)
+        if self.ref_storage == ProjectRefStorage.PROJECT_REFS:
+            self.refs.load(self.options)
+            self.junction_refs.load(self.options)
 
         # Parse shell options
         shell_options = _yaml.node_get(config, Mapping, 'shell')
@@ -341,12 +411,23 @@ class Project():
 
                 # Parse the host mount
                 path = _yaml.node_get(host_file_desc, str, 'path')
-                host_path = _yaml.node_get(host_file_desc, str, 'host_path', default_value='') or None
+                host_path = _yaml.node_get(host_file_desc, str, 'host_path', default_value=None)
                 optional = _yaml.node_get(host_file_desc, bool, 'optional', default_value=False)
                 mount = HostMount(path, host_path, optional)
 
             self._shell_host_files.append(mount)
 
+    # _assert_plugin_format()
+    #
+    # Helper to raise a PluginError if the loaded plugin is of a lesser version then
+    # the required version for this plugin
+    #
+    def _assert_plugin_format(self, plugin, version):
+        if plugin.BST_FORMAT_VERSION < version:
+            raise LoadError(LoadErrorReason.UNSUPPORTED_PLUGIN,
+                            "{}: Format version {} is too old for requested version {}"
+                            .format(plugin, plugin.BST_FORMAT_VERSION, version))
+
     # _store_origin()
     #
     # Helper function to store plugin origins
@@ -378,157 +459,6 @@ class Project():
                 origin_dict['path'] = os.path.join(self.directory, origin_dict['path'])
             destination.append(origin_dict)
 
-    # _list_workspaces()
-    #
-    # Generator function to enumerate workspaces.
-    #
-    # Yields:
-    #    A tuple in the following format: (element, path).
-    def _list_workspaces(self):
-        for element, _ in _yaml.node_items(self._workspaces):
-            yield (element, self._workspaces[element]["path"])
-
-    # _get_workspace()
-    #
-    # Get the path of the workspace source associated with the given
-    # element's source at the given index
-    #
-    # Args:
-    #    element (str) - The element name
-    #
-    # Returns:
-    #    None if no workspace is open, the path to the workspace
-    #    otherwise
-    #
-    def _get_workspace(self, element):
-        if element not in self._workspaces:
-            return None
-        return self._workspaces[element]["path"]
-
-    # _set_workspace()
-    #
-    # Set the path of the workspace associated with the given
-    # element's source at the given index
-    #
-    # Args:
-    #    element (str) - The element name
-    #    path (str) - The path to set the workspace to
-    #
-    def _set_workspace(self, element, path):
-        if element.name not in self._workspaces:
-            self._workspaces[element.name] = {}
-
-        self._workspaces[element.name]["path"] = path
-        element._set_source_workspaces(path)
-
-    # _delete_workspace()
-    #
-    # Remove the workspace from the workspace element. Note that this
-    # does *not* remove the workspace from the stored yaml
-    # configuration, call _save_workspace_config() afterwards.
-    #
-    # Args:
-    #    element (str) - The element name
-    #
-    def _delete_workspace(self, element):
-        del self._workspaces[element]
-
-    # _load_workspace_config()
-    #
-    # Load the workspace configuration and return a node containing
-    # all open workspaces for the project
-    #
-    # Returns:
-    #
-    #    A node containing a dict that assigns elements to their
-    #    workspaces. For example:
-    #
-    #        alpha.bst: /home/me/alpha
-    #        bravo.bst: /home/me/bravo
-    #
-    def _load_workspace_config(self):
-        os.makedirs(os.path.join(self.directory, ".bst"), exist_ok=True)
-        workspace_file = os.path.join(self.directory, ".bst", "workspaces.yml")
-        try:
-            open(workspace_file, "a").close()
-        except IOError as e:
-            raise LoadError(LoadErrorReason.MISSING_FILE,
-                            "Could not load workspace config: {}".format(e)) from e
-
-        return _yaml.load(workspace_file)
-
-    # _parse_workspace_config()
-    #
-    # If workspace config is in old-style format, i.e. it is using
-    # source-specific workspaces, try to convert it to element-specific
-    # workspaces.
-    #
-    # This method will rewrite workspace config, if it is in old format.
-    #
-    # Args:
-    #    workspaces (dict): current workspace config, usually output of _load_workspace_config()
-    #
-    # Returns:
-    #    (bool, dict) Whether the workspace config needs to be
-    #                 rewritten and extracted workspaces
-    #
-    # Raises: LoadError if there was a problem with the workspace config
-    #
-    def _parse_workspace_config(self, workspaces):
-        version = _yaml.node_get(workspaces, int, "format-version", default_value=0)
-
-        if version == 0:
-            # Pre-versioning format can be of two forms
-            for element, config in _yaml.node_items(workspaces):
-                if isinstance(config, str):
-                    pass
-
-                elif isinstance(config, dict):
-                    sources = list(_yaml.node_items(config))
-                    if len(sources) > 1:
-                        detail = "There are multiple workspaces open for '{}'.\n" + \
-                                 "This is not supported anymore.\n" + \
-                                 "Please remove this element from '{}'."
-                        raise LoadError(LoadErrorReason.INVALID_DATA,
-                                        detail.format(element,
-                                                      os.path.join(self.directory, ".bst", "workspaces.yml")))
-
-                    workspaces[element] = sources[0][1]
-
-                else:
-                    raise LoadError(LoadErrorReason.INVALID_DATA,
-                                    "Workspace config is in unexpected format.")
-
-            res = {
-                element: {"path": config}
-                for element, config in _yaml.node_items(workspaces)
-            }
-
-        elif version == BST_WORKSPACE_FORMAT_VERSION:
-            res = _yaml.node_get(workspaces, dict, "workspaces", default_value={})
-
-        else:
-            raise LoadError(LoadErrorReason.INVALID_DATA,
-                            "Workspace configuration format version {} not supported."
-                            "Your version of buildstream may be too old. Max supported version: {}"
-                            .format(version, BST_WORKSPACE_FORMAT_VERSION))
-
-        return res
-
-    # _save_workspace_config()
-    #
-    # Dump the current workspace element to the project configuration
-    # file. This makes any changes performed with _delete_workspace or
-    # _set_workspace permanent
-    #
-    def _save_workspace_config(self):
-        _yaml.dump(
-            {
-                "format-version": BST_WORKSPACE_FORMAT_VERSION,
-                "workspaces": _yaml.node_sanitize(self._workspaces)
-            },
-            os.path.join(self.directory, ".bst", "workspaces.yml"))
-
     def _extract_plugin_paths(self, node, name):
         if not node:
             return
@@ -536,25 +466,3 @@ class Project():
         for i in range(len(path_list)):
             path = _yaml.node_get(node, str, name, indices=[i])
             yield path
-
-    # _get_cache_key():
-    #
-    # Returns the cache key, calculating it if necessary
-    #
-    # Returns:
-    #    (str): A hex digest cache key for the Context
-    #
-    def _get_cache_key(self):
-        if self._cache_key is None:
-
-            # Anything that alters the build goes into the unique key
-            # (currently nothing here)
-            self._cache_key = _cachekey.generate_key({})
-
-        return self._cache_key
-
-    def _create_element(self, kind, artifacts, meta):
-        return self._element_factory.create(kind, self._context, self, artifacts, meta)
-
-    def _create_source(self, kind, meta):
-        return self._source_factory.create(kind, self._context, self, meta)
