@@ -20,6 +20,7 @@
 
 import sys
 import collections
+import string
 from copy import deepcopy
 from contextlib import ExitStack
 
@@ -304,6 +305,18 @@ def node_get_provenance(node, key=None, indices=None):
     return provenance
 
 
+# Helper to use utils.sentinel without unconditional utils import,
+# which causes issues for completion.
+#
+# Local private, but defined here because sphinx appears to break if
+# it's not defined before any functions calling it in default kwarg
+# values.
+#
+def _get_sentinel():
+    from .utils import _sentinel
+    return _sentinel
+
+
 # node_get()
 #
 # Fetches a value from a dictionary node and checks it for
@@ -325,10 +338,10 @@ def node_get_provenance(node, key=None, indices=None):
 # Note:
 #    Returned strings are stripped of leading and trailing whitespace
 #
-def node_get(node, expected_type, key, indices=None, default_value=None):
+def node_get(node, expected_type, key, indices=None, default_value=_get_sentinel()):
     value = node.get(key, default_value)
     provenance = node_get_provenance(node)
-    if value is None:
+    if value is _get_sentinel():
         raise LoadError(LoadErrorReason.INVALID_DATA,
                         "{}: Dictionary did not contain expected key '{}'".format(provenance, key))
 
@@ -340,6 +353,10 @@ def node_get(node, expected_type, key, indices=None, default_value=None):
             value = value[index]
             path += '[{:d}]'.format(index)
 
+    # We want to allow None as a valid value for any type
+    if value is None:
+        return None
+
     if not isinstance(value, expected_type):
         # Attempt basic conversions if possible, typically we want to
         # be able to specify numeric values and convert them to strings,
@@ -1015,3 +1032,49 @@ def list_final_assertions(values):
             node_final_assertions(value)
         elif isinstance(value, list):
             list_final_assertions(value)
+
+
+# assert_symbol_name()
+#
+# A helper function to check if a loaded string is a valid symbol
+# name and to raise a consistent LoadError if not. For strings which
+# are required to be symbols.
+#
+# Args:
+#    provenance (Provenance): The provenance of the loaded symbol, or None
+#    symbol_name (str): The loaded symbol name
+#    purpose (str): The purpose of the string, for an error message
+#    allow_dashes (bool): Whether dashes are allowed for this symbol
+#
+# Raises:
+#    LoadError: If the symbol_name is invalid
+#
+# Note that dashes are generally preferred for variable names and
+# usage in YAML, but things such as option names which will be
+# evaluated with jinja2 cannot use dashes.
+#
+def assert_symbol_name(provenance, symbol_name, purpose, *, allow_dashes=True):
+    valid_chars = string.digits + string.ascii_letters + '_'
+    if allow_dashes:
+        valid_chars += '-'
+
+    valid = True
+    if not symbol_name:
+        valid = False
+    elif any(x not in valid_chars for x in symbol_name):
+        valid = False
+    elif symbol_name[0] in string.digits:
+        valid = False
+
+    if not valid:
+        detail = "Symbol names must contain only alphanumeric characters, " + \
+                 "may not start with a digit, and may contain underscores"
+        if allow_dashes:
+            detail += " or dashes"
+
+        message = "Invalid symbol name for {}: '{}'".format(purpose, symbol_name)
+        if provenance is not None:
+            message = "{}: {}".format(provenance, message)
+
+        raise LoadError(LoadErrorReason.INVALID_SYMBOL_NAME,
+                        message, detail=detail)
