@@ -17,84 +17,67 @@
 #
 #  Authors:
 #        Tristan Van Berkom <tristan.vanberkom@codethink.co.uk>
-
-"""The BuildElement class is a convenience element one can derive from for
+"""
+BuildElement
+============
+The BuildElement class is a convenience element one can derive from for
 implementing the most common case of element.
 
 
-Description of assemble activities
-----------------------------------
-This element will perform the following steps to assemble an element:
+Abstract method implementations
+-------------------------------
 
-Stage dependencies
+Element.configure_sandbox()
+~~~~~~~~~~~~~~~~~~~~~~~~~~~
+In :func:`Element.configure_sandbox() <buildstream.element.Element.configure_sandbox>`,
+the BuildElement will ensure that the sandbox locations described by the ``%{build-root}``
+and ``%{install-root}`` variables are marked and will be mounted read-write for the
+:func:`assemble phase<buildstream.element.Element.configure_sandbox>`.
+
+The working directory for the sandbox will be configured to be the ``%{build-root}``,
+unless the ``%{command-subdir}`` variable is specified for the element in question,
+in which case the working directory will be configured as ``%{build-root}/%{command-subdir}``.
+
+
+Element.stage()
+~~~~~~~~~~~~~~~
+In :func:`Element.stage() <buildstream.element.Element.stage>`, the BuildElement
+will do the following operations:
+
+* Stage all the dependencies in the :func:`Scope.BUILD <buildstream.element.Scope.BUILD>`
+  scope into the sandbox root.
+
+* Run the integration commands for all staged dependencies using
+  :func:`Element.integrate() <buildstream.element.Element.integrate>`
+
+* Stage any Source on the given element to the ``%{build-root}`` location
+  inside the sandbox, using
+  :func:`Element.stage_sources() <buildstream.element.Element.integrate>`
+
+
+Element.prepare()
+~~~~~~~~~~~~~~~~~
+In :func:`Element.prepare() <buildstream.element.Element.prepare>`,
+the BuildElement will run ``configure-commands``, which are used to
+run one-off preparations that should not be repeated for a single
+build directory.
+
+
+Element.assemble()
 ~~~~~~~~~~~~~~~~~~
-The dependencies in the :func:`Scope.BUILD <buildstream.element.Scope.BUILD>`
-scope will be staged at the root of the sandbox
-
-Integrate dependencies
-~~~~~~~~~~~~~~~~~~~~~~
-The integration commands taken from the ``bst`` public domain of each dependency
-will be run in the sandbox to create and update caches. Typically ``ldconfig``
-among other things is run in this step.
-
-Stage sources
-~~~~~~~~~~~~~
-:mod:`Sources <buildstream.source>` are now staged according to their configuration
-into the ``%{build-root}`` directory (normally ``/buildstream/build``) inside the sandbox.
-
-Run commands
-~~~~~~~~~~~~
-Commands are now run in the sandbox.
-
-Commands are taken from the element configuration specified by the given
-:mod:`BuildElement <buildstream.buildelement>` subclass, which can in turn be
-overridden by the user in element declarations (``.bst`` files).
+In :func:`Element.assemble() <buildstream.element.Element.assemble>`, the
+BuildElement will proceed to run sandboxed commands which are expected to be
+found in the element configuration.
 
 Commands are run in the following order:
 
-* ``configure-commands``: Commands to configure how the element will build
 * ``build-commands``: Commands to build the element
 * ``install-commands``: Commands to install the results into ``%{install-root}``
 * ``strip-commands``: Commands to strip debugging symbols installed binaries
 
-Sometimes it is interesting to append or prepend commands to an existing
-command list without replacing it entirely, for this; array composition
-:ref:`prepend <format_directives_list_prepend>` and
-:ref:`append <format_directives_list_append>` directives can be used.
-
-**Example**
-
-.. code:: yaml
-
-  config:
-    configure-commands:
-      (<):
-      - echo "Do something before default configure-commands"
-
-**Working Directory**
-
-Note that by default the working directory is where the sources are staged in
-``%{build-root}``, but this can be overridden to build inside of a subdirectory
-of the build directory using the ``command-subdir`` variable in an element
-declaration. e.g.:
-
-.. code:: yaml
-
-  variables:
-    command-subdir: src
-
-The above fragment will cause all commands to be run in the ``src/`` subdirectory
-of the staged sources.
-
-
-Result collection
-~~~~~~~~~~~~~~~~~
-Finally, the resulting build *artifact* is collected from the the ``%{install-root}``
-directory (which is normally configured as ``/buildstream/install``) inside the sandbox.
-
-All build elements must install into the ``%{install-root}`` using whatever
-semantic the given build system provides to do this. E.g. for standard autotools
-packages we simply do ``make DESTDIR=%{install-root} install``.
+The result of the build is expected to end up in ``%{install-root}``, and
+as such; Element.assemble() method will return the ``%{install-root}`` for
+artifact collection purposes.
 """
 
 import os
@@ -120,9 +103,12 @@ _command_steps = ['configure-commands',
 
 class BuildElement(Element):
 
+    #############################################################
+    #             Abstract Method Implementations               #
+    #############################################################
     def configure(self, node):
 
-        self.commands = {}
+        self.__commands = {}
 
         # FIXME: Currently this forcefully validates configurations
         #        for all BuildElement subclasses so they are unable to
@@ -131,9 +117,9 @@ class BuildElement(Element):
 
         for command_name in _legacy_command_steps:
             if command_name in _command_steps:
-                self.commands[command_name] = self._get_commands(node, command_name)
+                self.__commands[command_name] = self.__get_commands(node, command_name)
             else:
-                self.commands[command_name] = []
+                self.__commands[command_name] = []
 
     def preflight(self):
         pass
@@ -141,15 +127,14 @@ class BuildElement(Element):
     def get_unique_key(self):
         dictionary = {}
 
-        for command_name, command_list in self.commands.items():
+        for command_name, command_list in self.__commands.items():
             dictionary[command_name] = command_list
 
         # Specifying notparallel for a given element effects the
         # cache key, while having the side effect of setting max-jobs to 1,
         # which is normally automatically resolved and does not effect
         # the cache key.
-        variables = self._get_variables()
-        if self.node_get_member(variables.variables, bool, 'notparallel', False):
+        if self.get_variable('notparallel'):
             dictionary['notparallel'] = True
 
         return dictionary
@@ -192,21 +177,13 @@ class BuildElement(Element):
 
         # Run commands
         for command_name in _command_steps:
-            commands = self.commands[command_name]
-            if not commands:
+            commands = self.__commands[command_name]
+            if not commands or command_name == 'configure-commands':
                 continue
 
             with self.timed_activity("Running {}".format(command_name)):
                 for cmd in commands:
-                    self.status("Running {}".format(command_name), detail=cmd)
-
-                    # Note the -e switch to 'sh' means to exit with an error
-                    # if any untested command fails.
-                    #
-                    exitcode = sandbox.run(['sh', '-c', '-e', cmd + '\n'],
-                                           SandboxFlags.ROOT_READ_ONLY)
-                    if exitcode != 0:
-                        raise ElementError("Command '{}' failed with exitcode {}".format(cmd, exitcode))
+                    self.__run_command(sandbox, cmd, command_name)
 
         # %{install-root}/%{build-root} should normally not be written
         # to - if an element later attempts to stage to a location
@@ -222,10 +199,29 @@ class BuildElement(Element):
                       "will not be allowed to stage to %{build-root}.")
 
         # Return the payload, this is configurable but is generally
-        # always the /buildstream/install directory
+        # always the /buildstream-install directory
         return self.get_variable('install-root')
 
-    def _get_commands(self, node, name):
+    def prepare(self, sandbox):
+        commands = self.__commands['configure-commands']
+        if commands:
+            for cmd in commands:
+                self.__run_command(sandbox, cmd, 'configure-commands')
+
+    def generate_script(self):
+        script = ""
+        for command_name in _command_steps:
+            commands = self.__commands[command_name]
+
+            for cmd in commands:
+                script += "(set -ex; {}\n) || exit 1\n".format(cmd)
+
+        return script
+
+    #############################################################
+    #                   Private Local Methods                   #
+    #############################################################
+    def __get_commands(self, node, name):
         list_node = self.node_get_member(node, list, name, [])
         commands = []
 
@@ -235,12 +231,14 @@ class BuildElement(Element):
 
         return commands
 
-    def generate_script(self):
-        script = ""
-        for command_name in _command_steps:
-            commands = self.commands[command_name]
+    def __run_command(self, sandbox, cmd, cmd_name):
+        with self.timed_activity("Running {}".format(cmd_name)):
+            self.status("Running {}".format(cmd_name), detail=cmd)
 
-            for cmd in commands:
-                script += "(set -ex; {}\n) || exit 1\n".format(cmd)
-
-        return script
+            # Note the -e switch to 'sh' means to exit with an error
+            # if any untested command fails.
+            #
+            exitcode = sandbox.run(['sh', '-c', '-e', cmd + '\n'],
+                                   SandboxFlags.ROOT_READ_ONLY)
+            if exitcode != 0:
+                raise ElementError("Command '{}' failed with exitcode {}".format(cmd, exitcode))
