@@ -19,9 +19,35 @@
 #        Tristan Van Berkom <tristan.vanberkom@codethink.co.uk>
 """
 Plugin
-=======
+======
 BuildStream supports third party plugins to define additional kinds of
-elements and sources.
+:mod:`Elements <buildstream.element>` and :mod:`Sources <buildstream.source>`.
+
+The common API is documented here, along with some information on how
+external plugin packages are structured.
+
+
+.. _core_plugin_abstract_methods:
+
+Abstract Methods
+----------------
+For both :mod:`Elements <buildstream.element>` and :mod:`Sources <buildstream.source>`,
+it is mandatory to implement the following abstract methods:
+
+* :func:`Plugin.configure() <buildstream.plugin.Plugin.configure>`
+
+  Loads the user provided configuration YAML for the given source or element
+
+* :func:`Plugin.preflight() <buildstream.plugin.Plugin.preflight>`
+
+  Early preflight checks allow plugins to bail out early with an error
+  in the case that it can predict that failure is inevitable.
+
+* :func:`Plugin.get_unique_key() <buildstream.plugin.Plugin.get_unique_key>`
+
+  Once all configuration has been loaded and preflight checks have passed,
+  this method is used to inform the core of a plugin's unique configuration.
+
 
 Plugin Structure
 ----------------
@@ -105,7 +131,7 @@ class Plugin():
     """The plugin's YAML format version
 
     This should be set to ``1`` the first time any new configuration
-    is understood by your :func:`~buildstream.plugin.Plugin.configure`
+    is understood by your :func:`Plugin.configure() <buildstream.plugin.Plugin.configure>`
     implementation and subsequently bumped every time your
     configuration is enhanced.
 
@@ -114,6 +140,10 @@ class Plugin():
        Plugins are expected to maintain backward compatibility
        in the format and configurations they expose. The versioning
        is intended to track availability of new features only.
+
+       For convenience, the format version for plugins maintained and
+       distributed with BuildStream are revisioned with BuildStream's
+       core format version :ref:`core format version <project_format_version>`.
     """
 
     def __init__(self, name, context, project, provenance, type_tag):
@@ -139,18 +169,6 @@ class Plugin():
         modulename = type(self).__module__
         self.__kind = modulename.split('.')[-1]
 
-        # Ugly special case to determine the minimum required version
-        # from the project.
-        if type_tag == 'element':
-            version = project._element_format_versions.get(self.get_kind(), 0)
-        else:
-            version = project._source_format_versions.get(self.get_kind(), 0)
-
-        # Raise PluginError on format version mismatch here
-        if self.BST_FORMAT_VERSION < version:
-            raise PluginError("{}: Format version {} is too old for requested version {}"
-                              .format(self, self.BST_FORMAT_VERSION, version))
-
         self.debug("Created: {}".format(self))
 
     def __del__(self):
@@ -165,6 +183,88 @@ class Plugin():
             typetag=self.__type_tag,
             provenance=self.__provenance)
 
+    #############################################################
+    #                      Abstract Methods                     #
+    #############################################################
+    def configure(self, node):
+        """Configure the Plugin from loaded configuration data
+
+        Args:
+           node (dict): The loaded configuration dictionary
+
+        Raises:
+           :class:`.SourceError`: If its a :class:`.Source` implementation
+           :class:`.ElementError`: If its an :class:`.Element` implementation
+
+        Plugin implementors should implement this method to read configuration
+        data and store it.
+
+        Plugins should use the :func:`Plugin.node_get_member() <buildstream.plugin.Plugin.node_get_member>`
+        and :func:`Plugin.node_get_list_element() <buildstream.plugin.Plugin.node_get_list_element>`
+        methods to fetch values from the passed `node`. This will ensure that a nice human readable error
+        message will be raised if the expected configuration is not found, indicating the filename,
+        line and column numbers.
+
+        Further the :func:`Plugin.node_validate() <buildstream.plugin.Plugin.node_validate>` method
+        should be used to ensure that the user has not specified keys in `node` which are unsupported
+        by the plugin.
+
+        .. note::
+
+           For Elements, when variable substitution is desirable, the
+           :func:`Element.node_subst_member() <buildstream.element.Element.node_subst_member>`
+           and :func:`Element.node_subst_list_element() <buildstream.element.Element.node_subst_list_element>`
+           methods can be used.
+        """
+        raise ImplError("{tag} plugin '{kind}' does not implement configure()".format(
+            tag=self.__type_tag, kind=self.get_kind()))
+
+    def preflight(self):
+        """Preflight Check
+
+        Raises:
+           :class:`.SourceError`: If its a :class:`.Source` implementation
+           :class:`.ElementError`: If its an :class:`.Element` implementation
+
+        This method is run after :func:`Plugin.configure() <buildstream.plugin.Plugin.configure>`
+        and after the pipeline is fully constructed.
+
+        Implementors should simply raise :class:`.SourceError` or :class:`.ElementError`
+        with an informative message in the case that the host environment is
+        unsuitable for operation.
+
+        Plugins which require host tools (only sources usually) should obtain
+        them with :func:`utils.get_host_tool() <buildstream.utils.get_host_tool>` which
+        will raise an error automatically informing the user that a host tool is needed.
+        """
+        raise ImplError("{tag} plugin '{kind}' does not implement preflight()".format(
+            tag=self.__type_tag, kind=self.get_kind()))
+
+    def get_unique_key(self):
+        """Return something which uniquely identifies the plugin input
+
+        Returns:
+           A string, list or dictionary which uniquely identifies the input
+
+        This is used to construct unique cache keys for elements and sources,
+        sources should return something which uniquely identifies the payload,
+        such as an sha256 sum of a tarball content.
+
+        Elements and Sources should implement this by collecting any configurations
+        which could possibly effect the output and return a dictionary of these settings.
+
+        For Sources, this is guaranteed to only be called if
+        :func:`Source.get_consistency() <buildstream.source.Source.get_consistency>`
+        has not returned :func:`Consistency.INCONSISTENT <buildstream.source.Consistency.INCONSISTENT>`
+        which is to say that the Source is expected to have an exact *ref* indicating
+        exactly what source is going to be staged.
+        """
+        raise ImplError("{tag} plugin '{kind}' does not implement get_unique_key()".format(
+            tag=self.__type_tag, kind=self.get_kind()))
+
+    #############################################################
+    #                       Public Methods                      #
+    #############################################################
     def get_kind(self):
         """Fetches the kind of this plugin
 
@@ -206,7 +306,7 @@ class Plugin():
         provenance = _yaml.node_get_provenance(node, key=member_name)
         return str(provenance)
 
-    def node_get_member(self, node, expected_type, member_name, default=None):
+    def node_get_member(self, node, expected_type, member_name, default=utils._sentinel):
         """Fetch the value of a node member, raising an error if the value is
         missing or incorrectly typed.
 
@@ -297,71 +397,6 @@ class Plugin():
         """
         return _yaml.node_get(node, expected_type, member_name, indices=indices)
 
-    def configure(self, node):
-        """Configure the Plugin from loaded configuration data
-
-        Args:
-           node (dict): The loaded configuration dictionary
-
-        Raises:
-           :class:`.SourceError`: If its a :class:`.Source` implementation
-           :class:`.ElementError`: If its an :class:`.Element` implementation
-           :class:`.LoadError`: If one of the *node* handling methods fail
-
-        Plugin implementors should implement this method to read configuration
-        data and store it. Use of the :func:`~buildstream.plugin.Plugin.node_get_member`
-        convenience method will ensure that a nice :class:`.LoadError` is triggered
-        whenever the YAML input configuration is faulty.
-
-        Implementations may raise :class:`.SourceError` or :class:`.ElementError` for other errors.
-
-        .. note::
-
-           During configure, logging is suppressed unless buildstream is run with
-           debugging output enabled.
-        """
-        raise ImplError("{tag} plugin '{kind}' does not implement configure()".format(
-            tag=self.__type_tag, kind=self.get_kind()))
-
-    def preflight(self):
-        """Preflight Check
-
-        Raises:
-           :class:`.SourceError`: If its a :class:`.Source` implementation
-           :class:`.ElementError`: If its an :class:`.Element` implementation
-           :class:`.ProgramNotFoundError`: If a required host tool is not found
-
-        This method is run after :func:`~buildstream.plugin.Plugin.configure` and
-        after the pipeline is fully constructed. :class:`.Element` plugins are free
-        to use the :func:`~buildstream.element.Element.dependencies` method and inspect
-        public data at this time.
-
-        Implementors should simply raise :class:`.SourceError` or :class:`.ElementError`
-        with an informative message in the case that the host environment is
-        unsuitable for operation.
-
-        Plugins which require host tools (only sources usually) should obtain
-        them with :func:`.utils.get_host_tool` which will raise
-        :class:`.ProgramNotFoundError` automatically.
-        """
-        raise ImplError("{tag} plugin '{kind}' does not implement preflight()".format(
-            tag=self.__type_tag, kind=self.get_kind()))
-
-    def get_unique_key(self):
-        """Return something which uniquely identifies the plugin input
-
-        Returns:
-           A string, list or dictionary which uniquely identifies the sources to use
-
-        This is used to construct unique cache keys for elements and sources,
-        sources should return something which uniquely identifies the payload,
-        such as an sha256 sum of a tarball content. Elements should implement
-        this by collecting any configurations which could possibly effect the
-        output and return a dictionary of these settings.
-        """
-        raise ImplError("{tag} plugin '{kind}' does not implement get_unique_key()".format(
-            tag=self.__type_tag, kind=self.get_kind()))
-
     def debug(self, brief, *, detail=None):
         """Print a debugging message
 
@@ -439,9 +474,9 @@ class Plugin():
               # This will raise SourceError on its own
               self.call(... command which takes time ...)
         """
-        with self.__context._timed_activity(activity_name,
-                                            detail=detail,
-                                            silent_nested=silent_nested):
+        with self.__context.timed_activity(activity_name,
+                                           detail=detail,
+                                           silent_nested=silent_nested):
             yield
 
     def call(self, *popenargs, fail=None, **kwargs):
@@ -615,7 +650,7 @@ class Plugin():
 
     def __message(self, message_type, brief, **kwargs):
         message = Message(self.__unique_id, message_type, brief, **kwargs)
-        self.__context._message(message)
+        self.__context.message(message)
 
     def __note_command(self, output, *popenargs, **kwargs):
         workdir = os.getcwd()
@@ -628,8 +663,8 @@ class Plugin():
 
     def _get_full_name(self):
         project = self.__project
-        if project._junction:
-            return '{}:{}'.format(project._junction.name, self.name)
+        if project.junction:
+            return '{}:{}'.format(project.junction.name, self.name)
         else:
             return self.name
 
@@ -660,13 +695,8 @@ __PLUGINS_TABLE = WeakValueDictionary()
 #    (Plugin): The plugin for the given ID, or None
 #
 def _plugin_lookup(unique_id):
-    try:
-        plugin = __PLUGINS_TABLE[unique_id]
-    except (AttributeError, KeyError) as e:
-        print("Could not find plugin with ID {}".format(unique_id))
-        raise
-
-    return plugin
+    assert unique_id in __PLUGINS_TABLE, "Could not find plugin with ID {}".format(unique_id)
+    return __PLUGINS_TABLE[unique_id]
 
 
 # No need for unregister, WeakValueDictionary() will remove entries
