@@ -115,6 +115,11 @@ from buildstream._pipeline import PipelineError
 
 # Element implementation for the 'junction' kind.
 class JunctionElement(Element):
+    # pylint: disable=attribute-defined-outside-init
+
+    # Junctions are not allowed any dependencies
+    BST_FORBID_BDEPENDS = True
+    BST_FORBID_RDEPENDS = True
 
     def configure(self, node):
         self.path = self.node_get_member(node, str, 'path', default='')
