@@ -20,6 +20,52 @@
 """
 Source
 ======
+
+
+.. _core_source_abstract_methods:
+
+Abstract Methods
+----------------
+For loading and configuration purposes, Sources must implement the
+:ref:`Plugin base class abstract methods <core_plugin_abstract_methods>`.
+
+Sources expose the following abstract methods. Unless explicitly mentioned,
+these methods are mandatory to implement.
+
+* :func:`Source.get_consistency() <buildstream.source.Source.get_consistency>`
+
+  Report the sources consistency state.
+
+* :func:`Source.load_ref() <buildstream.source.Source.load_ref>`
+
+  Load the ref from a specific YAML node
+
+* :func:`Source.get_ref() <buildstream.source.Source.get_ref>`
+
+  Fetch the source ref
+
+* :func:`Source.set_ref() <buildstream.source.Source.set_ref>`
+
+  Set a new ref explicitly
+
+* :func:`Source.track() <buildstream.source.Source.track>`
+
+  Automatically derive a new ref from a symbolic tracking branch
+
+* :func:`Source.fetch() <buildstream.source.Source.fetch>`
+
+  Fetch the actual payload for the currently set ref
+
+* :func:`Source.stage() <buildstream.source.Source.stage>`
+
+  Stage the sources for a given ref at a specified location
+
+* :func:`Source.init_workspace() <buildstream.source.Source.init_workspace>`
+
+  Stage sources in a local directory for use as a workspace.
+
+  **Optional**: If left unimplemented, this will default to calling
+  :func:`Source.stage() <buildstream.source.Source.stage>`
 """
 
 import os
@@ -28,7 +74,7 @@ from contextlib import contextmanager
 
 from . import Plugin
 from . import _yaml, utils
-from ._exceptions import BstError, ImplError, LoadError, LoadErrorReason, ErrorDomain
+from ._exceptions import BstError, ImplError, ErrorDomain
 from ._projectrefs import ProjectRefStorage
 
 
@@ -86,12 +132,9 @@ class Source(Plugin):
 
         self.__element_name = meta.element_name         # The name of the element owning this source
         self.__element_index = meta.element_index       # The index of the source in the owning element's source list
+        self.__element_kind = meta.element_kind         # The kind of the element owning this source
         self.__directory = meta.directory               # Staging relative directory
         self.__consistency = Consistency.INCONSISTENT   # Cached consistency state
-        self.__tracking = False                         # Source is scheduled to be tracked
-        self.__assemble_scheduled = False               # Source is scheduled to be assembled
-        self.__workspace = None                         # Directory of the currently active workspace
-        self.__workspace_key = None                     # Cached directory content hashes for workspaced source
 
         # Collect the composited element configuration and
         # ask the element to configure itself.
@@ -106,79 +149,9 @@ class Source(Plugin):
     should be checked for using node_validate().
     """
 
-    def __init_defaults(self):
-        if not self.__defaults_set:
-            project = self._get_project()
-            sources = project._sources
-            type(self).__defaults = sources.get(self.get_kind(), {})
-            type(self).__defaults_set = True
-
-    # This will resolve the final configuration to be handed
-    # off to source.configure()
-    #
-    def __extract_config(self, meta):
-        config = _yaml.node_get(self.__defaults, Mapping, 'config', default_value={})
-        config = _yaml.node_chain_copy(config)
-
-        _yaml.composite(config, meta.config)
-        _yaml.node_final_assertions(config)
-
-        return config
-
-    def get_mirror_directory(self):
-        """Fetches the directory where this source should store things
-
-        Returns:
-           (str): The directory belonging to this source
-        """
-
-        # Create the directory if it doesnt exist
-        context = self._get_context()
-        directory = os.path.join(context.sourcedir, self.get_kind())
-        os.makedirs(directory, exist_ok=True)
-        return directory
-
-    def translate_url(self, url):
-        """Translates the given url which may be specified with an alias
-        into a fully qualified url.
-
-        Args:
-           url (str): A url, which may be using an alias
-
-        Returns:
-           str: The fully qualified url, with aliases resolved
-        """
-        project = self._get_project()
-        return project.translate_url(url)
-
-    def get_project_directory(self):
-        """Fetch the project base directory
-
-        This is useful for sources which need to load resources
-        stored somewhere inside the project.
-
-        Returns:
-           str: The project base directory
-        """
-        project = self._get_project()
-        return project.directory
-
-    @contextmanager
-    def tempdir(self):
-        """Context manager for working in a temporary directory
-
-        Yields:
-           (str): A path to a temporary directory
-
-        This should be used by source plugins directly instead of the tempfile
-        module. This one will automatically cleanup in case of termination by
-        catching the signal before os._exit(). It will also use the 'mirror
-        directory' as expected for a source.
-        """
-        mirrordir = self.get_mirror_directory()
-        with utils._tempdir(dir=mirrordir) as tempdir:
-            yield tempdir
-
+    #############################################################
+    #                      Abstract Methods                     #
+    #############################################################
     def get_consistency(self):
         """Report whether the source has a resolved reference
 
@@ -193,6 +166,14 @@ class Source(Plugin):
         Args:
            node (dict): The YAML node to load the ref from
 
+        .. note::
+
+           The *ref* for the Source is expected to be read at
+           :func:`Plugin.configure() <buildstream.plugin.Plugin.configure>` time,
+           this will only be used for loading refs from alternative locations
+           than in the `element.bst` file where the given Source object has
+           been declared.
+
         *Since: 1.2*
         """
         raise ImplError("Source plugin '{}' does not implement load_ref()".format(self.get_kind()))
@@ -303,139 +284,126 @@ class Source(Plugin):
         """
         self.stage(directory)
 
+    #############################################################
+    #                       Public Methods                      #
+    #############################################################
+    def get_mirror_directory(self):
+        """Fetches the directory where this source should store things
+
+        Returns:
+           (str): The directory belonging to this source
+        """
+
+        # Create the directory if it doesnt exist
+        context = self._get_context()
+        directory = os.path.join(context.sourcedir, self.get_kind())
+        os.makedirs(directory, exist_ok=True)
+        return directory
+
+    def translate_url(self, url):
+        """Translates the given url which may be specified with an alias
+        into a fully qualified url.
+
+        Args:
+           url (str): A url, which may be using an alias
+
+        Returns:
+           str: The fully qualified url, with aliases resolved
+        """
+        project = self._get_project()
+        return project.translate_url(url)
+
+    def get_project_directory(self):
+        """Fetch the project base directory
+
+        This is useful for sources which need to load resources
+        stored somewhere inside the project.
+
+        Returns:
+           str: The project base directory
+        """
+        project = self._get_project()
+        return project.directory
+
+    @contextmanager
+    def tempdir(self):
+        """Context manager for working in a temporary directory
+
+        Yields:
+           (str): A path to a temporary directory
+
+        This should be used by source plugins directly instead of the tempfile
+        module. This one will automatically cleanup in case of termination by
+        catching the signal before os._exit(). It will also use the 'mirror
+        directory' as expected for a source.
+        """
+        mirrordir = self.get_mirror_directory()
+        with utils._tempdir(dir=mirrordir) as tempdir:
+            yield tempdir
+
     #############################################################
     #            Private Methods used in BuildStream            #
     #############################################################
 
+    # Wrapper around preflight() method
+    #
+    def _preflight(self):
+        try:
+            self.preflight()
+        except BstError as e:
+            # Prepend provenance to the error
+            raise SourceError("{}: {}".format(self, e), reason=e.reason) from e
+
     # Update cached consistency for a source
     #
     # This must be called whenever the state of a source may have changed.
     #
     def _update_state(self):
-        if self.__tracking:
-            return
 
         if self.__consistency < Consistency.CACHED:
 
             # Source consistency interrogations are silent.
             context = self._get_context()
-            with context._silence():
+            with context.silence():
                 self.__consistency = self.get_consistency()
 
-            if self._has_workspace() and \
-               self.__consistency > Consistency.INCONSISTENT:
-
-                # A workspace is considered inconsistent in the case
-                # that it's directory went missing
-                #
-                fullpath = self._get_workspace_path()
-                if not os.path.exists(fullpath):
-                    self.__consistency = Consistency.INCONSISTENT
-
     # Return cached consistency
     #
     def _get_consistency(self):
         return self.__consistency
 
-    # Return the absolute path of the element's workspace
-    #
-    def _get_workspace_path(self):
-        return os.path.join(self.get_project_directory(), self.__workspace)
-
-    # Mark a source as scheduled to be tracked
-    #
-    # This is used across the pipeline in sessions where the
-    # source in question are going to be tracked. This is important
-    # as it will prevent depending elements from producing cache
-    # keys until the source is RESOLVED and also prevent depending
-    # elements from being assembled until the source is CACHED.
-    #
-    def _schedule_tracking(self):
-        self.__tracking = True
-
-    # _schedule_assemble():
-    #
-    # This is called in the main process before the element is assembled
-    # in a subprocess.
-    #
-    def _schedule_assemble(self):
-        assert not self.__assemble_scheduled
-        self.__assemble_scheduled = True
-
-        # Invalidate workspace key as the build modifies the workspace directory
-        self.__workspace_key = None
-
-    # _assemble_done():
-    #
-    # This is called in the main process after the element has been assembled
-    # in a subprocess.
-    #
-    def _assemble_done(self):
-        assert self.__assemble_scheduled
-        self.__assemble_scheduled = False
-
-    # _stable():
-    #
-    # Unstable sources are mounted read/write and thus cannot produce a
-    # (stable) cache key before the build is complete.
-    #
-    def _stable(self):
-        # Source directory is modified by workspace build process
-        return not (self._has_workspace() and self.__assemble_scheduled)
-
     # Wrapper function around plugin provided fetch method
     #
     def _fetch(self):
         self.fetch()
 
-    # Return the path where this source should be staged under given directory
-    def _get_staging_path(self, directory):
-        if self.__directory is not None:
-            directory = os.path.join(directory, self.__directory.lstrip(os.sep))
-        return directory
-
-    # Ensures a fully constructed path and returns it
-    def _ensure_directory(self, directory):
-        directory = self._get_staging_path(directory)
-        try:
-            os.makedirs(directory, exist_ok=True)
-        except OSError as e:
-            raise SourceError("Failed to create staging directory: {}"
-                              .format(e),
-                              reason="ensure-stage-dir-fail") from e
-        return directory
-
     # Wrapper for stage() api which gives the source
     # plugin a fully constructed path considering the
     # 'directory' option
     #
     def _stage(self, directory):
-        staging_directory = self._ensure_directory(directory)
+        staging_directory = self.__ensure_directory(directory)
 
-        if self._has_workspace():
-            self._stage_workspace(staging_directory)
-        else:
-            self.stage(staging_directory)
+        self.stage(staging_directory)
 
     # Wrapper for init_workspace()
     def _init_workspace(self, directory):
-        directory = self._ensure_directory(directory)
+        directory = self.__ensure_directory(directory)
 
         self.init_workspace(directory)
 
+    # _get_unique_key():
+    #
     # Wrapper for get_unique_key() api
     #
-    # This adds any core attributes to the key and
-    # also calculates something different if workspaces
-    # are active.
+    # Args:
+    #    include_source (bool): Whether to include the delegated source key
     #
-    def _get_unique_key(self):
+    def _get_unique_key(self, include_source):
         key = {}
 
         key['directory'] = self.__directory
-        if self._has_workspace():
-            key['workspace'] = self._get_workspace_key()
-        else:
+        if include_source:
             key['unique'] = self.get_unique_key()
 
         return key
@@ -456,10 +424,22 @@ class Source(Plugin):
         # the ref, regardless of whether the original has changed.
         self.set_ref(ref, node)
 
-        self.__tracking = False
-
         return changed
 
+    # _project_refs():
+    #
+    # Gets the appropriate ProjectRefs object for this source,
+    # which depends on whether the owning element is a junction
+    #
+    # Args:
+    #    project (Project): The project to check
+    #
+    def _project_refs(self, project):
+        element_kind = self.__element_kind
+        if element_kind == 'junction':
+            return project.junction_refs
+        return project.refs
+
     # _load_ref():
     #
     # Loads the ref for the said source.
@@ -481,7 +461,7 @@ class Source(Plugin):
     def _load_ref(self):
         context = self._get_context()
         project = self._get_project()
-        toplevel = context._get_toplevel_project()
+        toplevel = context.get_toplevel_project()
         redundant_ref = None
 
         element_name = self.__element_name
@@ -496,8 +476,9 @@ class Source(Plugin):
                                   reason="unsupported-load-ref") from e
 
         # If the main project overrides the ref, use the override
-        if project is not toplevel and toplevel._ref_storage == ProjectRefStorage.PROJECT_REFS:
-            ref_node = toplevel.refs.lookup_ref(project.name, element_name, element_idx)
+        if project is not toplevel and toplevel.ref_storage == ProjectRefStorage.PROJECT_REFS:
+            refs = self._project_refs(toplevel)
+            ref_node = refs.lookup_ref(project.name, element_name, element_idx)
             if ref_node is not None:
                 do_load_ref(ref_node)
 
@@ -506,7 +487,7 @@ class Source(Plugin):
         # violate the rule of refs being either in project.refs or in
         # the elements themselves.
         #
-        elif project._ref_storage == ProjectRefStorage.PROJECT_REFS:
+        elif project.ref_storage == ProjectRefStorage.PROJECT_REFS:
 
             # First warn if there is a ref already loaded, and reset it
             redundant_ref = self.get_ref()
@@ -514,7 +495,8 @@ class Source(Plugin):
                 self.set_ref(None, {})
 
             # Try to load the ref
-            ref_node = project.refs.lookup_ref(project.name, element_name, element_idx)
+            refs = self._project_refs(project)
+            ref_node = refs.lookup_ref(project.name, element_name, element_idx)
             if ref_node is not None:
                 do_load_ref(ref_node)
 
@@ -539,7 +521,8 @@ class Source(Plugin):
 
         context = self._get_context()
         project = self._get_project()
-        toplevel = context._get_toplevel_project()
+        toplevel = context.get_toplevel_project()
+        toplevel_refs = self._project_refs(toplevel)
         provenance = self._get_provenance()
 
         element_name = self.__element_name
@@ -549,13 +532,13 @@ class Source(Plugin):
         # Step 1 - Obtain the node
         #
         if project is toplevel:
-            if toplevel._ref_storage == ProjectRefStorage.PROJECT_REFS:
-                node = toplevel.refs.lookup_ref(project.name, element_name, element_idx, write=True)
+            if toplevel.ref_storage == ProjectRefStorage.PROJECT_REFS:
+                node = toplevel_refs.lookup_ref(project.name, element_name, element_idx, write=True)
             else:
                 node = provenance.node
         else:
-            if toplevel._ref_storage == ProjectRefStorage.PROJECT_REFS:
-                node = toplevel.refs.lookup_ref(project.name, element_name, element_idx, write=True)
+            if toplevel.ref_storage == ProjectRefStorage.PROJECT_REFS:
+                node = toplevel_refs.lookup_ref(project.name, element_name, element_idx, write=True)
             else:
                 node = {}
 
@@ -576,8 +559,8 @@ class Source(Plugin):
         # Step 3 - Apply the change in project data
         #
         if project is toplevel:
-            if toplevel._ref_storage == ProjectRefStorage.PROJECT_REFS:
-                do_save_refs(toplevel.refs)
+            if toplevel.ref_storage == ProjectRefStorage.PROJECT_REFS:
+                do_save_refs(toplevel_refs)
             else:
                 # Save the ref in the originating file
                 #
@@ -589,8 +572,8 @@ class Source(Plugin):
                                       .format(self, provenance.filename, e),
                                       reason="save-ref-error") from e
         else:
-            if toplevel._ref_storage == ProjectRefStorage.PROJECT_REFS:
-                do_save_refs(toplevel.refs)
+            if toplevel.ref_storage == ProjectRefStorage.PROJECT_REFS:
+                do_save_refs(toplevel_refs)
             else:
                 self.warn("{}: Not persisting new reference in junctioned project".format(self))
 
@@ -608,87 +591,42 @@ class Source(Plugin):
 
         if current_ref != new_ref:
             self.info("Found new revision: {}".format(new_ref))
-            if self._has_workspace():
-                detail = "This source has an open workspace.\n" \
-                    + "To start using the new reference, please close the existing workspace."
-                self.warn("Updated reference will be ignored as source has open workspace", detail=detail)
 
         return new_ref
 
-    # Set the current workspace directory
+    #############################################################
+    #                   Local Private Methods                   #
+    #############################################################
+
+    # Ensures a fully constructed path and returns it
+    def __ensure_directory(self, directory):
+
+        if self.__directory is not None:
+            directory = os.path.join(directory, self.__directory.lstrip(os.sep))
+
+        try:
+            os.makedirs(directory, exist_ok=True)
+        except OSError as e:
+            raise SourceError("Failed to create staging directory: {}"
+                              .format(e),
+                              reason="ensure-stage-dir-fail") from e
+        return directory
+
+    def __init_defaults(self):
+        if not self.__defaults_set:
+            project = self._get_project()
+            sources = project.source_overrides
+            type(self).__defaults = sources.get(self.get_kind(), {})
+            type(self).__defaults_set = True
+
+    # This will resolve the final configuration to be handed
+    # off to source.configure()
     #
-    # Note that this invalidate the workspace key.
-    #
-    def _set_workspace(self, directory):
-        self.__workspace = directory
-        self.__workspace_key = None
+    def __extract_config(self, meta):
+        config = _yaml.node_get(self.__defaults, Mapping, 'config', default_value={})
+        config = _yaml.node_chain_copy(config)
 
-    # Return the current workspace directory
-    def _get_workspace(self):
-        return self.__workspace
+        _yaml.composite(config, meta.config)
+        _yaml.node_final_assertions(config)
 
-    # Delete the workspace
-    #
-    # Note that this invalidate the workspace key.
-    #
-    def _del_workspace(self):
-        self.__workspace = None
-        self.__workspace_key = None
-
-    # Whether the source has a set workspace
-    #
-    def _has_workspace(self):
-        return self.__workspace is not None
-
-    # Stage the workspace
-    #
-    def _stage_workspace(self, directory):
-        fullpath = self._get_workspace_path()
-
-        with self.timed_activity("Staging local files at {}".format(self.__workspace)):
-            if os.path.isdir(fullpath):
-                utils.copy_files(fullpath, directory)
-            else:
-                destfile = os.path.join(directory, os.path.basename(self.__workspace))
-                utils.safe_copy(fullpath, destfile)
-
-    # Get a unique key for the workspace
-    #
-    # Note that to avoid re-traversing the file system if this function is
-    # called multiple times, the workspace key is cached. You can still force a
-    # new calculation to happen by setting the 'recalculate' flag.
-    #
-    def _get_workspace_key(self, recalculate=False):
-        assert not self.__assemble_scheduled
-
-        if recalculate or self.__workspace_key is None:
-            fullpath = self._get_workspace_path()
-
-            # Get a list of tuples of the the project relative paths and fullpaths
-            if os.path.isdir(fullpath):
-                filelist = utils.list_relative_paths(fullpath)
-                filelist = [(relpath, os.path.join(fullpath, relpath)) for relpath in filelist]
-            else:
-                filelist = [(self.__workspace, fullpath)]
-
-            # Return a list of (relative filename, sha256 digest) tuples, a sorted list
-            # has already been returned by list_relative_paths()
-            self.__workspace_key = [(relpath, _unique_key(fullpath)) for relpath, fullpath in filelist]
-
-        return self.__workspace_key
-
-
-# Get the sha256 sum for the content of a file
-def _unique_key(filename):
-
-    # If it's a directory, just return 0 string
-    if os.path.isdir(filename):
-        return "0"
-    elif os.path.islink(filename):
-        return "1"
-
-    try:
-        return utils.sha256sum(filename)
-    except FileNotFoundError as e:
-        raise LoadError(LoadErrorReason.MISSING_FILE,
-                        "Failed loading workspace. Did you remove the workspace directory? {}".format(e))
+        return config
