@@ -39,7 +39,6 @@ from contextlib import contextmanager
 
 import psutil
 
-from . import __version__
 from . import _signals
 from ._exceptions import BstError, ErrorDomain
 
@@ -85,6 +84,18 @@ class FileListResult():
         self.files_written = []
         """List of files that were written."""
 
+    def combine(self, other):
+        """Create a new FileListResult that contains the results of both.
+        """
+        ret = FileListResult()
+
+        ret.overwritten = self.overwritten + other.overwritten
+        ret.ignored = self.ignored + other.ignored
+        ret.failed_attributes = self.failed_attributes + other.failed_attributes
+        ret.files_written = self.files_written + other.files_written
+
+        return ret
+
 
 def list_relative_paths(directory):
     """A generator for walking directory relative paths
@@ -457,6 +468,8 @@ def get_bst_version():
        (int): The major version
        (int): The minor version
     """
+    # Import this only conditionally, it's not resolved at bash complete time
+    from . import __version__
     versions = __version__.split('.')[:2]
 
     return (int(versions[0]), int(versions[1]))
@@ -515,11 +528,28 @@ def save_file_atomic(filename, mode='w', *, buffering=-1, encoding=None,
             # This operation is atomic, at least on platforms we care about:
             # https://bugs.python.org/issue8828
             os.replace(tempname, filename)
-    except Exception as e:
+    except Exception:
         cleanup_tempfile()
         raise
 
 
+# A sentinel to be used as a default argument for functions that need
+# to distinguish between a kwarg set to None and an unset kwarg.
+_sentinel = object()
+
+# Main process pid
+_main_pid = os.getpid()
+
+
+# _is_main_process()
+#
+# Return whether we are in the main process or not.
+#
+def _is_main_process():
+    assert _main_pid is not None
+    return os.getpid() == _main_pid
+
+
 # Recursively remove directories, ignoring file permissions as much as
 # possible.
 def _force_rmtree(rootpath, **kwargs):
@@ -656,7 +686,7 @@ def _process_list(srcdir, destdir, filelist, actionfunc, result,
             if ignore_missing:
                 continue
             else:
-                raise UtilError("Source file is missing: {}".format(srcpath))
+                raise UtilError("Source file is missing: {}".format(srcpath)) from e
 
         if stat.S_ISDIR(mode):
             # Ensure directory exists in destination
