@@ -2,9 +2,11 @@ import os
 import pytest
 from tests.testutils import cli, create_repo, ALL_REPO_KINDS
 
-from buildstream._exceptions import ErrorDomain
+from buildstream._exceptions import ErrorDomain, LoadErrorReason
 from buildstream import _yaml
 
+from . import configure_project, generate_junction
+
 # Project directory
 TOP_DIR = os.path.dirname(os.path.realpath(__file__))
 DATA_DIR = os.path.join(TOP_DIR, 'project')
@@ -23,12 +25,6 @@ def generate_element(repo, element_path, dep_name=None):
     _yaml.dump(element, element_path)
 
 
-def configure_project(path, config):
-    config['name'] = 'test'
-    config['element-path'] = 'elements'
-    _yaml.dump(config, os.path.join(path, 'project.conf'))
-
-
 @pytest.mark.datafiles(DATA_DIR)
 @pytest.mark.parametrize("ref_storage", [('inline'), ('project.refs')])
 @pytest.mark.parametrize("kind", [(kind) for kind in ALL_REPO_KINDS])
@@ -120,6 +116,46 @@ def test_track_recurse(cli, tmpdir, datafiles, kind):
     assert cli.get_element_state(project, element_target_name) == 'waiting'
 
 
+@pytest.mark.datafiles(DATA_DIR)
+def test_track_single(cli, tmpdir, datafiles):
+    project = os.path.join(datafiles.dirname, datafiles.basename)
+    dev_files_path = os.path.join(project, 'files', 'dev-files')
+    element_path = os.path.join(project, 'elements')
+    element_dep_name = 'track-test-dep.bst'
+    element_target_name = 'track-test-target.bst'
+
+    # Create our repo object of the given source type with
+    # the dev files, and then collect the initial ref.
+    #
+    repo = create_repo('git', str(tmpdir))
+    ref = repo.create(dev_files_path)
+
+    # Write out our test targets
+    generate_element(repo, os.path.join(element_path, element_dep_name))
+    generate_element(repo, os.path.join(element_path, element_target_name),
+                     dep_name=element_dep_name)
+
+    # Assert that tracking is needed for both elements
+    assert cli.get_element_state(project, element_dep_name) == 'no reference'
+    assert cli.get_element_state(project, element_target_name) == 'no reference'
+
+    # Now first try to track only one element
+    result = cli.run(project=project, args=[
+        'track', '--deps', 'none',
+        element_target_name])
+    result.assert_success()
+
+    # And now fetch it
+    result = cli.run(project=project, args=[
+        'fetch', '--deps', 'none',
+        element_target_name])
+    result.assert_success()
+
+    # Assert that the dependency is waiting and the target has still never been tracked
+    assert cli.get_element_state(project, element_dep_name) == 'no reference'
+    assert cli.get_element_state(project, element_target_name) == 'waiting'
+
+
 @pytest.mark.datafiles(DATA_DIR)
 @pytest.mark.parametrize("kind", [(kind) for kind in ALL_REPO_KINDS])
 def test_track_recurse_except(cli, tmpdir, datafiles, kind):
@@ -219,8 +255,9 @@ def test_track_optional(cli, tmpdir, datafiles, ref_storage):
 
 
 @pytest.mark.datafiles(os.path.join(TOP_DIR, 'track-cross-junction'))
+@pytest.mark.parametrize("cross_junction", [('cross'), ('nocross')])
 @pytest.mark.parametrize("ref_storage", [('inline'), ('project.refs')])
-def test_track_cross_junction(cli, tmpdir, datafiles, ref_storage):
+def test_track_cross_junction(cli, tmpdir, datafiles, cross_junction, ref_storage):
     project = os.path.join(datafiles.dirname, datafiles.basename)
     dev_files_path = os.path.join(project, 'files')
     target_path = os.path.join(project, 'target.bst')
@@ -271,14 +308,27 @@ def test_track_cross_junction(cli, tmpdir, datafiles, ref_storage):
     assert get_subproject_element_state() == 'no reference'
 
     # Track recursively across the junction
-    result = cli.run(project=project, args=['track', '--deps', 'all', 'target.bst'])
+    args = ['track', '--deps', 'all']
+    if cross_junction == 'cross':
+        args += ['--cross-junctions']
+    args += ['target.bst']
+
+    result = cli.run(project=project, args=args)
 
     if ref_storage == 'inline':
-        #
-        # Cross junction tracking is not allowed when the toplevel project
-        # is using inline ref storage.
-        #
-        result.assert_main_error(ErrorDomain.PIPELINE, 'untrackable-sources')
+
+        if cross_junction == 'cross':
+            #
+            # Cross junction tracking is not allowed when the toplevel project
+            # is using inline ref storage.
+            #
+            result.assert_main_error(ErrorDomain.PIPELINE, 'untrackable-sources')
+        else:
+            #
+            # No cross juction tracking was requested
+            #
+            result.assert_success()
+            assert get_subproject_element_state() == 'no reference'
     else:
         #
         # Tracking is allowed with project.refs ref storage
@@ -286,9 +336,12 @@ def test_track_cross_junction(cli, tmpdir, datafiles, ref_storage):
         result.assert_success()
 
         #
-        # Assert that we now have a ref for the subproject element
+        # If cross junction tracking was enabled, we should now be buildable
         #
-        assert get_subproject_element_state() == 'buildable'
+        if cross_junction == 'cross':
+            assert get_subproject_element_state() == 'buildable'
+        else:
+            assert get_subproject_element_state() == 'no reference'
 
 
 @pytest.mark.datafiles(os.path.join(TOP_DIR, 'consistencyerror'))
@@ -310,3 +363,77 @@ def test_track_consistency_bug(cli, tmpdir, datafiles):
 
     # We expect BuildStream to fail gracefully, with no recorded exception.
     result.assert_main_error(ErrorDomain.PIPELINE, None)
+
+
+@pytest.mark.datafiles(DATA_DIR)
+@pytest.mark.parametrize("ref_storage", [('inline'), ('project.refs')])
+def test_inconsistent_junction(cli, tmpdir, datafiles, ref_storage):
+    project = os.path.join(datafiles.dirname, datafiles.basename)
+    subproject_path = os.path.join(project, 'files', 'sub-project')
+    junction_path = os.path.join(project, 'elements', 'junction.bst')
+    element_path = os.path.join(project, 'elements', 'junction-dep.bst')
+
+    configure_project(project, {
+        'ref-storage': ref_storage
+    })
+
+    # Create a repo to hold the subproject and generate a junction element for it
+    generate_junction(tmpdir, subproject_path, junction_path, store_ref=False)
+
+    # Create a stack element to depend on a cross junction element
+    #
+    element = {
+        'kind': 'stack',
+        'depends': [
+            {
+                'junction': 'junction.bst',
+                'filename': 'import-etc.bst'
+            }
+        ]
+    }
+    _yaml.dump(element, element_path)
+
+    # Now try to track it, this will bail with the appropriate error
+    # informing the user to track the junction first
+    result = cli.run(project=project, args=['track', 'junction-dep.bst'])
+    result.assert_main_error(ErrorDomain.LOAD, LoadErrorReason.SUBPROJECT_INCONSISTENT)
+
+
+@pytest.mark.datafiles(DATA_DIR)
+@pytest.mark.parametrize("ref_storage", [('inline'), ('project.refs')])
+def test_junction_element(cli, tmpdir, datafiles, ref_storage):
+    project = os.path.join(datafiles.dirname, datafiles.basename)
+    subproject_path = os.path.join(project, 'files', 'sub-project')
+    junction_path = os.path.join(project, 'elements', 'junction.bst')
+    element_path = os.path.join(project, 'elements', 'junction-dep.bst')
+
+    configure_project(project, {
+        'ref-storage': ref_storage
+    })
+
+    # Create a repo to hold the subproject and generate a junction element for it
+    generate_junction(tmpdir, subproject_path, junction_path, store_ref=False)
+
+    # Create a stack element to depend on a cross junction element
+    #
+    element = {
+        'kind': 'stack',
+        'depends': [
+            {
+                'junction': 'junction.bst',
+                'filename': 'import-etc.bst'
+            }
+        ]
+    }
+    _yaml.dump(element, element_path)
+
+    # First demonstrate that showing the pipeline yields an error
+    result = cli.run(project=project, args=['show', 'junction-dep.bst'])
+    result.assert_main_error(ErrorDomain.LOAD, LoadErrorReason.SUBPROJECT_INCONSISTENT)
+
+    # Now track the junction itself
+    result = cli.run(project=project, args=['track', 'junction.bst'])
+    result.assert_success()
+
+    # Now assert element state (via bst show under the hood) of the dep again
+    assert cli.get_element_state(project, 'junction-dep.bst') == 'waiting'
